local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")

function Capitalist:InitScanStats()
	self:DebugEntry("InitScanStats")
--	Capitalist.hit = 0
--	Capitalist.miss = 0
	self:DebugMisc(time())

	for productName, data in pairs(self.list) do
		data.ownAuctions = 0
		data.otherAuctions = {}
		data.cost = nil
		data.bestPrice = nil
	end

	self:DebugMisc(time())
--	self:DebugMisc(Capitalist.hit)
--	self:DebugMisc(Capitalist.miss)
	self:DebugExit("InitScanStats")
end

function Capitalist:EnsureProductCost(productName)
	local data = self.list[productName]
	if (not data.cost) then
		-- calculate cost of the product
		data.cost = self:CalculateSkillCopperCost(data.profession, data.skillName, false)

		-- add auction house cut to the price and minProfit
		data.cost = self.db.profile.minProfit + data.cost / 0.95
		-- TODO: math change - the formula above looks incorrect in two ways:
		-- 1. the auction house cut is 5% not 5.263% (i.e. 1/0.95)
		-- 2. if user specifies minprofit that should be net after AH cut
		-- data.cost = (self.db.profile.minProfit + data.cost) * 1.05
	end
end

function Capitalist:EnsureBestPrice(productName)
	self:EnsureProductCost(productName)
	local data = self.list[productName]
	if (not data.bestPrice) then
		-- calculate maximum price
		data.bestPrice = data.cost * (1 + self.db.profile.margin / 100)
		if (data.bestPrice - data.cost < self.db.profile.minMargin) then
			data.bestPrice = data.cost + self.db.profile.minMargin
		end
		if (data.bestPrice - data.cost > self.db.profile.maxMargin) then
			data.bestPrice = data.cost + self.db.profile.maxMargin
		end
		data.bestPrice = math.floor(data.bestPrice) -- kills problems with float numbers
	end
end

function Capitalist:StartScan()
	self:DebugEntry("StartScan")

	if (self.scanBuckets and (#self.scanBuckets > 0)) then
		self.ScanBucketIndex = 1
		self.ScanBucketPage = 0
		self.ScanBucketCountBatch = nil
		self.ScanBucketCountTotal = nil
		self.ScanBucketCountRead = 0
	
		self:QueryBucket()

		self.auctionUI.lastScanLabel:Hide()
		self.auctionUI.progressLabel:Show()
		self.auctionUI.progressLabel:SetText(L["Scanning"].."...")
		self.auctionUI.progress:Show()
		self.auctionUI.progress:SetWidth(1)
		self.auctionUI.scanButton:Disable()
		self.auctionUI.postButton:Disable()
	
		self.mode = "scanning"

		self:InitScanStats()
	
		self.updateHandler:Show()
	end

	self:DebugExit("StartScan")
end

function Capitalist:QueryBucket()
	self:DebugEntry("QueryBucket")
	self.ScanQueryPending = nil
	self.ScanAnalyzed = nil

	if (self.ScanBucketPage == 0) then
		self.auctionUI.progressLabel:SetText(string.format(L["Scanning %d/%d"], self.ScanBucketIndex, #self.scanBuckets))
		self:DebugMisc(self.scanBuckets[self.ScanBucketIndex])
	end

	self:DebugMisc(self.scanBuckets[self.ScanBucketIndex]..", page "..self.ScanBucketPage)
	QueryAuctionItems(self.scanBuckets[self.ScanBucketIndex],nil,nil,0,0,0,self.ScanBucketPage,0,0,nil)
	self:DebugExit("QueryBucket")
end

function Capitalist:OnUpdateScanning()
	--self:DebugEntry("OnUpdateScanning")

	if (not self.InAuctionItemListUpdate) then
		if (self.ScanQueryPending) then
			if (CanSendAuctionQuery()) then
				self:QueryBucket()
			end
		end
	end

	--self:DebugExit("OnUpdateScanning")
end

function Capitalist:OnAuctionItemListUpdate()
	self:DebugEntry("OnAuctionItemListUpdate")
	self.InAuctionItemListUpdate = true

	self.ScanBucketCountBatch, self.ScanBucketCountTotal = GetNumAuctionItems("list")
	self:AnalyzeAuctionPage()
	self.ScanBucketCountRead = self.ScanBucketCountRead + self.ScanBucketCountBatch

	self:DebugMisc("T "..self.ScanBucketCountTotal.." B "..self.ScanBucketCountBatch.." R "..self.ScanBucketCountRead)

	if ((self.ScanBucketCountBatch > 0) and
		(self.ScanBucketCountRead < self.ScanBucketCountTotal)) then

		self.ScanBucketPage = self.ScanBucketPage + 1
		self.ScanQueryPending = true
	else
		self.ScanBucketCountRead = 0
		self.ScanBucketIndex = self.ScanBucketIndex + 1
		if (self.ScanBucketIndex > #self.scanBuckets) then
			self:ScanFinished()
		else
			self.ScanBucketPage = 0
			self.ScanQueryPending = true
		end
	end

	self.auctionUI.progress:SetWidth(200 * self.ScanBucketIndex / #self.scanBuckets)

	self.InAuctionItemListUpdate = false
	self:DebugExit("OnAuctionItemListUpdate")
end

function Capitalist:AnalyzeAuctionPage()
	self:DebugEntry("AnalyzeAuctionPage")
	
	self.ScanAnalyzed = true

	local batch, count = GetNumAuctionItems("list")
	self:DebugMisc(batch.."<batch count>"..count)

	for i = 1, self.ScanBucketCountBatch do
		local name, _, count, _, _, _, _, _, buyoutPrice, _, _, owner, _  = GetAuctionItemInfo("list", i)

		-- REVIEW: round up instead of down?
		local buyoutUnitPrice = buyoutPrice / count

		-- TODO: Add concept of "my" w/r/t "same account" auctions

		-- REVIEW: Correction for two glyphs named "Glyph of Death Coil"
		if (name == "Glyph of Death Coil") then
			local auctionLink = GetAuctionItemLink("list", i)
			if (auctionLink) then
				local itemIdString = string.match(auctionLink, "Hitem:(%d+)")
				local itemIdNumber = tonumber(itemIdString)
			
				-- item 45804, spell 64266 - DK
				if (itemIdNumber == 45804) then
					name = name.." <Death Knight>"
				end
				-- item 42457, spell 57261 - Warlock
				if (itemIdNumber == 42457) then
					name = name.." <Warlock>"
				end
			end
		end

		local v = self.list[name]
		if (v) then
			if (self:ProductIsEnabled(v.profession, v.skillName)) then
				self:EnsureBestPrice(name)
				if (owner == character) then
					v.ownAuctions = v.ownAuctions + 1
					if (v.buyoutSelf == nil or buyoutUnitPrice < v.buyoutSelf) then
						v.buyoutSelf = buyoutUnitPrice
					end
				else
					if (buyoutUnitPrice > 0) then
						table.insert(v.otherAuctions, buyoutUnitPrice)
						if ((v.buyoutOthers == nil) or
							(buyoutUnitPrice < v.buyoutOthers)) then
							v.buyoutOthers = buyoutUnitPrice
							v.otherOwner = owner
					end
					end
				end
			end
		end
	end
	
	self:DebugExit("AnalyzeAuctionPage")
end

function Capitalist:ScanFinished()
	self:DebugEntry("ScanFinished")
	self.updateHandler:Hide()

	self.auctionUI.lastScanLabel:SetText(L["Last Scan: "]..date("%H:%M"))
	self.auctionUI.lastScanLabel:Show()
	self.auctionUI.progress:Hide()
	self.auctionUI.progressLabel:Hide()
	self.auctionUI.scanButton:Enable()
	
	-- consistency check
	for i = 1, GetNumAuctionItems("owner") do
		local name, _, _, _, _, _, _, _, _, _, _, _, saleStatus = GetAuctionItemInfo("owner", i);

		-- REVIEW: Correction for two glyphs named "Glyph of Death Coil"
		if (name == "Glyph of Death Coil") then
			local auctionLink = GetAuctionItemLink("owner", i)
			if (auctionLink) then
				local itemIdString = string.match(auctionLink, "Hitem:(%d+)")
				local itemIdNumber = tonumber(itemIdString)
			
				-- item 45804, spell 64266 - DK
				if (itemIdNumber == 45804) then
					name = name.." <Death Knight>"
				end
				-- item 42457, spell 57261 - Warlock
				if (itemIdNumber == 42457) then
					name = name.." <Warlock>"
				end
			end
		end

		if ((saleStatus == 0) and self.list[name]) then
			local v = self.list[name]
			if (self:ProductIsEnabled(v.profession, v.skillName)) then
				self.list[name].ownAuctions = self.list[name].ownAuctions - 1
			end
		end
	end
	
	for k, v in pairs(self.list) do
		if (v.ownAuctions ~= 0) then
			self:DebugMisc(k)
			self:DebugMisc(v.profession)
			self:DebugMisc(v.skillName)
			self:DebugMisc(v.ownAuctions)
--print(k)
--print(v.profession)
--print(v.skillName)
--print(v.ownAuctions)

			self.auctionUI.lastScanLabel:SetText(L["Last Scan: failed"])
			self.validScan = false
		end
	end

	-- calculating optimal prices by undercutting competition
	for k, v in pairs(self.list) do
		if (self:ProductIsEnabled(v.profession, v.skillName)) then
			for j = 1, #v.otherAuctions do
				if (v.otherAuctions[j] <= v.bestPrice) then
					if (v.otherAuctions[j] > v.cost) then
						v.bestPrice = v.otherAuctions[j] - 1
					end
				end
			end
		end
	end

	self.validScan = true
	self:UpdateAuctionUI()
	self.auctionUI.postButton:Enable()
	self:CheckForCancel()
	self:DebugExit("ScanFinished")
end

function Capitalist:CheckForCancel()
	self.toCancel = {}

	for i = 1, GetNumAuctionItems("owner") do
		local name, _, count, _, _, _, _, _, buyoutPrice, bidAmount, _, _, _ = GetAuctionItemInfo("owner", i);

		-- REVIEW: round up instead of down?
		local buyoutUnitPrice = buyoutPrice / count

		-- REVIEW: Correction for two glyphs named "Glyph of Death Coil"
		if (name == "Glyph of Death Coil") then
			local auctionLink = GetAuctionItemLink("owner", i)
			if (auctionLink) then
				local itemIdString = string.match(auctionLink, "Hitem:(%d+)")
				local itemIdNumber = tonumber(itemIdString)
			
				-- item 45804, spell 64266 - DK
				if (itemIdNumber == 45804) then
					name = name.." <Death Knight>"
				end
				-- item 42457, spell 57261 - Warlock
				if (itemIdNumber == 42457) then
					name = name.." <Warlock>"
				end
			end
		end

		if (self.list[name] and (bidAmount == 0)) then
			self:EnsureBestPrice(name)
			if (buyoutUnitPrice > self.list[name].bestPrice and self.db.profile.cancelYellow) then
				table.insert(self.toCancel, i)
			elseif (buyoutUnitPrice < self.list[name].bestPrice and self.db.profile.cancelBlue) then
				table.insert(self.toCancel, i)
			elseif (self.db.profile.cancelDuration and GetAuctionItemTimeLeft("owner", i) ~= 4) then
				table.insert(self.toCancel, i)
			end
		end
	end

	if (#self.toCancel > 0) then
		self.auctionUI.cancelButton:Enable()
	else
		self.auctionUI.cancelButton:Disable()
	end
end

function Capitalist:CancelAuctions()
	self:DebugEntry("CancelAuctions")
	for i = 1, table.getn(self.toCancel) do
		CancelAuctionWithPrompt(self.toCancel[i])
	end
	self.toCancel = {}
	self.auctionUI.cancelButton:Disable()
	self:DebugExit("CancelAuctions")
end
